/**********************************************************************
  sample01.c:

   This program transforms input data values to output data values.
   It can be executed with an arbitrary number of processes.
   Its input and output should match the corresponding values in c06fxf3.r. 
   This program does not require any input parameter.

***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <openfft.h>

int main(int argc, char* argv[])
{ 
  dcomplex *Rhor, *Rhok;
  int numprocs,myid;
  int N1,N2,N3,measure_time,print_memory,My_Max_NumGrid,My_NumGrid_In,My_NumGrid_Out;
  int i,j,k,l;
  dcomplex ***Input;
  double factor;
  int My_Index_In[6],My_Index_Out[6];

  /* MPI */
  MPI_Init(&argc, &argv); 
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);

  /* Set global input */

  N1 = 2;
  N2 = 3;
  N3 = 4;

  Input = (dcomplex***)malloc(sizeof(dcomplex**)*N1);
  for(i=0;i<N1;i++){
    Input[i]=(dcomplex**)malloc(sizeof(dcomplex*)*N2);
    for(j=0;j<N2;j++){
      Input[i][j]=(dcomplex*)malloc(sizeof(dcomplex)*N3);
    }
  }

  Input[0][0][0].r=1.000; Input[0][0][0].i=0.000;
  Input[0][0][1].r=0.999; Input[0][0][1].i=-0.040;
  Input[0][0][2].r=0.987; Input[0][0][2].i=-0.159;
  Input[0][0][3].r=0.936; Input[0][0][3].i=-0.352;
  Input[0][1][0].r=0.994; Input[0][1][0].i=-0.111;
  Input[0][1][1].r=0.989; Input[0][1][1].i=-0.151;
  Input[0][1][2].r=0.963; Input[0][1][2].i=-0.268;
  Input[0][1][3].r=0.891; Input[0][1][3].i=-0.454;
  Input[0][2][0].r=0.903; Input[0][2][0].i=-0.430;
  Input[0][2][1].r=0.885; Input[0][2][1].i=-0.466;
  Input[0][2][2].r=0.823; Input[0][2][2].i=-0.568;
  Input[0][2][3].r=0.694; Input[0][2][3].i=-0.720;
  Input[1][0][0].r=0.500; Input[1][0][0].i=0.500;
  Input[1][0][1].r=0.499; Input[1][0][1].i=0.040;
  Input[1][0][2].r=0.487; Input[1][0][2].i=0.159;
  Input[1][0][3].r=0.436; Input[1][0][3].i=0.352;
  Input[1][1][0].r=0.494; Input[1][1][0].i=0.111;
  Input[1][1][1].r=0.489; Input[1][1][1].i=0.151;
  Input[1][1][2].r=0.463; Input[1][1][2].i=0.268;
  Input[1][1][3].r=0.391; Input[1][1][3].i=0.454;
  Input[1][2][0].r=0.403; Input[1][2][0].i=0.430;
  Input[1][2][1].r=0.385; Input[1][2][1].i=0.466;
  Input[1][2][2].r=0.323; Input[1][2][2].i=0.568;
  Input[1][2][3].r=0.194; Input[1][2][3].i=0.720;

  /* Set whether to use the timing and print memory functions of OpenFFT or not. Default=0 (not use) */

  measure_time = 1;
  print_memory = 0;

  /* Initialize OpenFFT */ 

  openfft_initialize(N1,N2,N3,measure_time,print_memory,
		     &My_Max_NumGrid,&My_NumGrid_In,My_Index_In,&My_NumGrid_Out,My_Index_Out);

  /* Allocate local input and output arrays */

  Rhor = (dcomplex*)malloc(sizeof(dcomplex)*My_Max_NumGrid); 
  Rhok = (dcomplex*)malloc(sizeof(dcomplex)*My_Max_NumGrid); 

  /* Set local input */

  MPI_Barrier(MPI_COMM_WORLD);

  printf("myid=%d, Input in the ABC(XYZ) order with %d grid points from (A=%d,B=%d,C=%d) to (A=%d,B=%d,C=%d)\n",myid,
	 My_NumGrid_In,My_Index_In[0],My_Index_In[1],My_Index_In[2],
	 My_Index_In[3],My_Index_In[4],My_Index_In[5]);

  l=0;

  if(My_Index_In[0]==My_Index_In[3]){
    i=My_Index_In[0];
    for(j=My_Index_In[1];j<=My_Index_In[4];j++){
      for(k=My_Index_In[2];k<=My_Index_In[5];k++){
	Rhor[l].r = Input[i][j][k].r;  
	Rhor[l].i = Input[i][j][k].i;  
	l++;
      }
    }
  }
  else if(My_Index_In[0]<My_Index_In[3]){
    for(i=My_Index_In[0];i<=My_Index_In[3];i++){
      if(i==My_Index_In[0]){
	for(j=My_Index_In[1];j<N2;j++){
	  for(k=My_Index_In[2];k<=My_Index_In[5];k++){
	    Rhor[l].r = Input[i][j][k].r;  
	    Rhor[l].i = Input[i][j][k].i;  
	    l++;
	  }
	}
      }
      else if(My_Index_In[0]<i && i<My_Index_In[3]){
	for(j=0;j<N2;j++){
	  for(k=My_Index_In[2];k<=My_Index_In[5];k++){
	    Rhor[l].r = Input[i][j][k].r;  
	    Rhor[l].i = Input[i][j][k].i;  
	    l++;
	  }
	}
      }
      else if(i==My_Index_In[3]){
	for(j=0;j<=My_Index_In[4];j++){
	  for(k=My_Index_In[2];k<=My_Index_In[5];k++){
	    Rhor[l].r = Input[i][j][k].r;  
	    Rhor[l].i = Input[i][j][k].i;  
	    l++;
	  }
	}
      }
    }
  }

  /* FFT transform */

  openfft_execute(Rhor, Rhok);

  /* Get local output and print in global form */
  factor = sqrt(N1*N2*N3);
  MPI_Barrier(MPI_COMM_WORLD);

  printf("myid=%d, Output in the CBA(ZYX) order with %d grid points from (C=%d,B=%d,A=%d) to (C=%d,B=%d,A=%d)\n",myid,
	 My_NumGrid_Out,My_Index_Out[0],My_Index_Out[1],My_Index_Out[2],
	 My_Index_Out[3],My_Index_Out[4],My_Index_Out[5]);


  l=0;

  if(My_Index_Out[0]==My_Index_Out[3]){
    i=My_Index_Out[0];
    for(j=My_Index_Out[1];j<=My_Index_Out[4];j++){
      for(k=My_Index_Out[2];k<=My_Index_Out[5];k++){
	printf("myid=%2d  Output[%d][%d][%d] Real: = %3.3f  Imag = %3.3f\n",myid,k,j,i,Rhok[l].r/factor,Rhok[l].i/factor);
	l++;
      }
    }
  }
  else if(My_Index_Out[0]<My_Index_Out[3]){
    for(i=My_Index_Out[0];i<=My_Index_Out[3];i++){
      if(i==My_Index_Out[0]){
	for(j=My_Index_Out[1];j<N2;j++){
	  for(k=My_Index_Out[2];k<=My_Index_Out[5];k++){
	    printf("myid=%2d  Output[%d][%d][%d]: Real = %3.3f  Imag = %3.3f\n",myid,k,j,i,Rhok[l].r/factor,Rhok[l].i/factor);
	    l++;
	  }
	}
      }
      else if(My_Index_Out[0]<i && i<My_Index_Out[3]){
	for(j=0;j<N2;j++){
	  for(k=My_Index_Out[2];k<=My_Index_Out[5];k++){
	    printf("myid=%2d  Output[%d][%d][%d]: Real = %3.3f  Imag = %3.3f\n",myid,k,j,i,Rhok[l].r/factor,Rhok[l].i/factor);
	    l++;
	  }
	}
      }
      else if(i==My_Index_Out[3]){
	for(j=0;j<=My_Index_Out[4];j++){
	  for(k=My_Index_Out[2];k<=My_Index_Out[5];k++){
	    printf("myid=%2d  Output[%d][%d][%d]: Real = %3.3f  Imag = %3.3f\n",myid,k,j,i,Rhok[l].r/factor,Rhok[l].i/factor);
	    l++;
	  }
	}
      }
    }
  }

  /* free arrays */

  free(Rhor);
  free(Rhok);

  /* finalize OpenFFT */

  openfft_finalize(MPI_COMM_WORLD);

  MPI_Finalize();

}

