/**********************************************************************
  openfft_execute is used to transform the data 
     Input: Rhor
     Output: Rhok
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "openfft.h"
#define numprocsk 5000

double openfft_execute(dcomplex *Rhor, dcomplex *Rhok) 
{
  int i,BN_AB,BN_CB,BN_CA,gp,NN_S,NN_R;
  dcomplex *array0,*array1;
  int *count_snd,*count_rcv,*displ_snd,*displ_rcv; 
  int numprocs,myid,tag=999,ID,IDS,IDR,done;
  double Stime_proc,Etime_proc,time0,time1,timeC,timeF;
  MPI_Status stat;
  MPI_Request request;
  MPI_Status *stat_send;
  MPI_Status *stat_recv;
  MPI_Request *request_send;
  MPI_Request *request_recv;
  MPI_Comm mpi_comm_level1 = MPI_COMM_WORLD;

  done = 0;
  time0 = 0.0;
  timeC = 0.0;
  timeF = 0.0;

  /* MPI */
  MPI_Comm_size(mpi_comm_level1,&numprocs);
  MPI_Comm_rank(mpi_comm_level1,&myid);

  /*------------------ FFT along the C-axis in the AB partition ------------------*/

  if (OFFT_measure_time==1){
    openfft_dtime(&Stime_proc);
    openfft_dtime(&time0);
  }
   
  for (BN_AB=0; BN_AB<My_NumGrid_AB; BN_AB+=OFFT_Ngrid3){

    for (i=0; i<OFFT_Ngrid3; i++){
      OFFT_in[i][0] = Rhor[BN_AB+i].r;
      OFFT_in[i][1] = Rhor[BN_AB+i].i;
    }

    fftw_execute(OFFT_p_C);

    for (i=0; i<OFFT_Ngrid3; i++){
      Rhor[BN_AB+i].r = OFFT_out[i][0];
      Rhor[BN_AB+i].i = OFFT_out[i][1];
    }
  }

  if (OFFT_measure_time==1){
    openfft_dtime(&time1);
    timeF += time1 - time0;
  }

  /*------------------ MPI: AB to CA partitions ------------------*/

  array0 = (dcomplex*)malloc(sizeof(dcomplex)*GP_AB2CA_S[NN_AB2CA_S]); 
  array1 = (dcomplex*)malloc(sizeof(dcomplex)*GP_AB2CA_R[NN_AB2CA_R]); 

#ifdef kcomp

if(numprocs < numprocsk){
  done = 1;

  count_snd = (int*)malloc(sizeof(int)*numprocs);
  count_rcv = (int*)malloc(sizeof(int)*numprocs);
  displ_snd = (int*)malloc(sizeof(int)*numprocs);
  displ_rcv = (int*)malloc(sizeof(int)*numprocs);

  for(i=0;i<numprocs;i++){
    count_snd[i] = 0;
    count_rcv[i] = 0;
    displ_snd[i] = 0;
    displ_rcv[i] = 0;
  }

  for (ID=0; ID<NN_AB2CA_S; ID++){

    IDS = ID_NN_AB2CA_S[ID];
    gp = GP_AB2CA_S[ID];

    count_snd[IDS] = Num_Snd_Grid_AB2CA[IDS]*2;
    displ_snd[IDS] = gp*2; 

    for (i=0; i<Num_Snd_Grid_AB2CA[IDS]; i++){
      BN_AB = Index_Snd_Grid_AB2CA[IDS][i];
      array0[gp+i].r = Rhor[BN_AB].r;
      array0[gp+i].i = Rhor[BN_AB].i;
    }     
  }

  for (ID=0; ID<NN_AB2CA_R; ID++){

    IDR = ID_NN_AB2CA_R[ID];
    gp = GP_AB2CA_R[ID];

    count_rcv[IDR] = Num_Rcv_Grid_AB2CA[IDR]*2;
    displ_rcv[IDR] = gp*2; 
  }


  if (OFFT_measure_time==1) openfft_dtime(&time0);

  MPI_Alltoallv(array0, count_snd, displ_snd, MPI_DOUBLE,
                       array1, count_rcv, displ_rcv, MPI_DOUBLE, mpi_comm_level1);

  if (OFFT_measure_time==1){
    openfft_dtime(&time1);
    timeC += time1 - time0;
  }

  /* copy them to Rhok */

  for (ID=0; ID<NN_AB2CA_R; ID++){

    IDR = ID_NN_AB2CA_R[ID];
    gp = GP_AB2CA_R[ID];

    for (i=0; i<Num_Rcv_Grid_AB2CA[IDR]; i++){
      BN_CA = Index_Rcv_Grid_AB2CA[IDR][i];
      Rhok[BN_CA].r = array1[gp+i].r;
      Rhok[BN_CA].i = array1[gp+i].i;
    }
  }

  free(count_snd);
  free(count_rcv);
  free(displ_snd);
  free(displ_rcv); 

}

#endif

if(!done){

  request_send = malloc(sizeof(MPI_Request)*NN_AB2CA_S);
  request_recv = malloc(sizeof(MPI_Request)*NN_AB2CA_R);
  stat_send = malloc(sizeof(MPI_Status)*NN_AB2CA_S);
  stat_recv = malloc(sizeof(MPI_Status)*NN_AB2CA_R);

  NN_S = 0;
  NN_R = 0;

  if (OFFT_measure_time==1) openfft_dtime(&time0);
   
  /* MPI_Irecv */

  for (ID=0; ID<NN_AB2CA_R; ID++){

    IDR = ID_NN_AB2CA_R[ID];
    gp = GP_AB2CA_R[ID];

    if (IDR!=myid){ 
      MPI_Irecv( &array1[gp], Num_Rcv_Grid_AB2CA[IDR]*2, 
                 MPI_DOUBLE, IDR, tag, mpi_comm_level1, &request_recv[NN_R]);
      NN_R++; 
    }
  }

  /* MPI_Isend */

  for (ID=0; ID<NN_AB2CA_S; ID++){

    IDS = ID_NN_AB2CA_S[ID];
    gp = GP_AB2CA_S[ID];

    if (IDS!=myid){

      for (i=0; i<Num_Snd_Grid_AB2CA[IDS]; i++){
        BN_AB = Index_Snd_Grid_AB2CA[IDS][i];
        array0[gp+i].r = Rhor[BN_AB].r;
        array0[gp+i].i = Rhor[BN_AB].i;
      }     

      MPI_Isend( &array0[gp], Num_Snd_Grid_AB2CA[IDS]*2, 
    	         MPI_DOUBLE, IDS, tag, mpi_comm_level1, &request_send[NN_S]);
      NN_S++;  
    }
  }
  
  /* MPI_Waitall */

  if (NN_S!=0) MPI_Waitall(NN_S,request_send,stat_send);
  if (NN_R!=0) MPI_Waitall(NN_R,request_recv,stat_recv);

  if (OFFT_measure_time==1){
    openfft_dtime(&time1);
    timeC += time1 - time0;
  }

  /* copy them to Rhok */

  for (ID=0; ID<NN_AB2CA_R; ID++){

    IDR = ID_NN_AB2CA_R[ID];
    gp = GP_AB2CA_R[ID];

    if (IDR!=myid){
      for (i=0; i<Num_Rcv_Grid_AB2CA[IDR]; i++){
	BN_CA = Index_Rcv_Grid_AB2CA[IDR][i];
	Rhok[BN_CA].r = array1[gp+i].r;
	Rhok[BN_CA].i = array1[gp+i].i;
      }
    }

    else{
      for (i=0; i<Num_Snd_Grid_AB2CA[IDR]; i++){
	BN_AB = Index_Snd_Grid_AB2CA[IDR][i];
	BN_CA = Index_Rcv_Grid_AB2CA[IDR][i];
	Rhok[BN_CA].r = Rhor[BN_AB].r;
	Rhok[BN_CA].i = Rhor[BN_AB].i;
      }
    }
  }

  free(request_send);
  free(request_recv);
  free(stat_send);
  free(stat_recv);

}

  free(array0);
  free(array1);



  /*------------------ FFT along the B-axis in the CA partition ------------------*/

  if (OFFT_measure_time==1) openfft_dtime(&time0);
    
  for (BN_CA=0; BN_CA<My_NumGrid_CA; BN_CA+=OFFT_Ngrid2){

    for (i=0; i<OFFT_Ngrid2; i++){
      OFFT_in[i][0] = Rhok[BN_CA+i].r;
      OFFT_in[i][1] = Rhok[BN_CA+i].i;
    }

    fftw_execute(OFFT_p_B);

    for (i=0; i<OFFT_Ngrid2; i++){
      Rhok[BN_CA+i].r = OFFT_out[i][0];
      Rhok[BN_CA+i].i = OFFT_out[i][1];
    }
  }

  if (OFFT_measure_time==1){
    openfft_dtime(&time1);
    timeF += time1 - time0;
  }

  /*------------------ MPI: CA to CB partitions ------------------*/

  done = 0;

  array0 = (dcomplex*)malloc(sizeof(dcomplex)*GP_CA2CB_S[NN_CA2CB_S]); 
  array1 = (dcomplex*)malloc(sizeof(dcomplex)*GP_CA2CB_R[NN_CA2CB_R]); 

#ifdef kcomp

if(numprocs < numprocsk){
  done = 1;

  count_snd = (int*)malloc(sizeof(int)*numprocs);
  count_rcv = (int*)malloc(sizeof(int)*numprocs);
  displ_snd = (int*)malloc(sizeof(int)*numprocs);
  displ_rcv = (int*)malloc(sizeof(int)*numprocs);

  for(i=0;i<numprocs;i++){
    count_snd[i] = 0;
    count_rcv[i] = 0;
    displ_snd[i] = 0;
    displ_rcv[i] = 0;
  }

  for (ID=0; ID<NN_CA2CB_S; ID++){

    IDS = ID_NN_CA2CB_S[ID];
    gp = GP_CA2CB_S[ID];

    count_snd[IDS] = Num_Snd_Grid_CA2CB[IDS]*2;
    displ_snd[IDS] = gp*2; 

    for (i=0; i<Num_Snd_Grid_CA2CB[IDS]; i++){
      BN_CA = Index_Snd_Grid_CA2CB[IDS][i];
      array0[gp+i].r = Rhok[BN_CA].r;
      array0[gp+i].i = Rhok[BN_CA].i;
    }     
  }

  for (ID=0; ID<NN_CA2CB_R; ID++){

    IDR = ID_NN_CA2CB_R[ID];
    gp = GP_CA2CB_R[ID];

    count_rcv[IDR] = Num_Rcv_Grid_CA2CB[IDR]*2;
    displ_rcv[IDR] = gp*2; 
  }

  if (OFFT_measure_time==1) openfft_dtime(&time0);

  MPI_Alltoallv(array0, count_snd, displ_snd, MPI_DOUBLE,
                       array1, count_rcv, displ_rcv, MPI_DOUBLE, mpi_comm_level1);

  if (OFFT_measure_time==1){
    openfft_dtime(&time1);
    timeC += time1 - time0;
  }

  /* copy them to Rhor */

  for (ID=0; ID<NN_CA2CB_R; ID++){

    IDR = ID_NN_CA2CB_R[ID];
    gp = GP_CA2CB_R[ID];

    for (i=0; i<Num_Rcv_Grid_CA2CB[IDR]; i++){
      BN_CB = Index_Rcv_Grid_CA2CB[IDR][i];
      Rhor[BN_CB].r = array1[gp+i].r;
      Rhor[BN_CB].i = array1[gp+i].i;
    }
  }

  free(count_snd);
  free(count_rcv);
  free(displ_snd);
  free(displ_rcv); 

}

#endif

if(!done){

  request_send = malloc(sizeof(MPI_Request)*NN_CA2CB_S);
  request_recv = malloc(sizeof(MPI_Request)*NN_CA2CB_R);
  stat_send = malloc(sizeof(MPI_Status)*NN_CA2CB_S);
  stat_recv = malloc(sizeof(MPI_Status)*NN_CA2CB_R);

  NN_S = 0;
  NN_R = 0;

  if (OFFT_measure_time==1) openfft_dtime(&time0);

  /* MPI_Irecv */

  for (ID=0; ID<NN_CA2CB_R; ID++){

    IDR = ID_NN_CA2CB_R[ID];
    gp = GP_CA2CB_R[ID];

    if (IDR!=myid){ 
      MPI_Irecv( &array1[gp], Num_Rcv_Grid_CA2CB[IDR]*2, 
                 MPI_DOUBLE, IDR, tag, mpi_comm_level1, &request_recv[NN_R]);
      NN_R++; 
    }
  }

  /* MPI_Isend */

  for (ID=0; ID<NN_CA2CB_S; ID++){

    IDS = ID_NN_CA2CB_S[ID];
    gp = GP_CA2CB_S[ID];

    if (IDS!=myid){

      for (i=0; i<Num_Snd_Grid_CA2CB[IDS]; i++){
        BN_CA = Index_Snd_Grid_CA2CB[IDS][i];
        array0[gp+i].r = Rhok[BN_CA].r;
        array0[gp+i].i = Rhok[BN_CA].i;
      }     

      MPI_Isend( &array0[gp], Num_Snd_Grid_CA2CB[IDS]*2, 
  	         MPI_DOUBLE, IDS, tag, mpi_comm_level1, &request_send[NN_S]);
      NN_S++;  
    }
  }

  /* MPI_Waitall */

  if (NN_S!=0) MPI_Waitall(NN_S,request_send,stat_send);
  if (NN_R!=0) MPI_Waitall(NN_R,request_recv,stat_recv);

  if (OFFT_measure_time==1){
    openfft_dtime(&time1);
    timeC += time1 - time0;
  }

  /* copy them to Rhor */

  for (ID=0; ID<NN_CA2CB_R; ID++){

    IDR = ID_NN_CA2CB_R[ID];
    gp = GP_CA2CB_R[ID];

    if (IDR!=myid){
      for (i=0; i<Num_Rcv_Grid_CA2CB[IDR]; i++){
        BN_CB = Index_Rcv_Grid_CA2CB[IDR][i];
        Rhor[BN_CB].r = array1[gp+i].r;
        Rhor[BN_CB].i = array1[gp+i].i;
      }
    }

    else{
      for (i=0; i<Num_Snd_Grid_CA2CB[IDR]; i++){
	BN_CA = Index_Snd_Grid_CA2CB[IDR][i];
	BN_CB = Index_Rcv_Grid_CA2CB[IDR][i];
	Rhor[BN_CB].r = Rhok[BN_CA].r;
	Rhor[BN_CB].i = Rhok[BN_CA].i;
      }
    }
  }

  free(request_send);
  free(request_recv);
  free(stat_send);
  free(stat_recv);

}

  free(array0);
  free(array1);


  /*------------------ FFT along the A-axis in the CB partition ------------------*/

  if (OFFT_measure_time==1) openfft_dtime(&time0);
   
  for (BN_CB=0; BN_CB<My_NumGrid_CB; BN_CB+=OFFT_Ngrid1){

    for (i=0; i<OFFT_Ngrid1; i++){
      OFFT_in[i][0] = Rhor[BN_CB+i].r;
      OFFT_in[i][1] = Rhor[BN_CB+i].i;
    }

    fftw_execute(OFFT_p_A);

    for (i=0; i<OFFT_Ngrid1; i++){
      Rhok[BN_CB+i].r = OFFT_out[i][0];
      Rhok[BN_CB+i].i = OFFT_out[i][1];
    }
  }

  if (OFFT_measure_time==1){
    openfft_dtime(&time1);
    timeF += time1 - time0;

    openfft_dtime(&Etime_proc);
    time0 = Etime_proc -Stime_proc;

    OFFT_timers[0] = timeF;
    OFFT_timers[1] = timeC;
    OFFT_timers[2] = time0 - timeF - timeC;
    OFFT_timers[3] = time0;
  }

  return time0;
}

